# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Builds Creative Maps network."""

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

from __future__ import annotations

import base64
import json
import logging
import operator
import os
import textwrap
from collections.abc import Mapping, Sequence
from io import BytesIO
from typing import Any, TypeAlias, TypedDict

import numpy as np
import pydantic
import smart_open
from garf_core import report
from media_similarity import media_similarity_service
from media_tagging import media, tagging_result
from PIL import Image, ImageDraw, ImageFont, ImageOps

MetricInfo: TypeAlias = dict[str, int | float]
Info: TypeAlias = dict[str, int | float | str | list[str] | None]


class MediaInfo(pydantic.BaseModel):
  """Contains extra information on a given medium."""

  media_path: str
  media_name: str
  info: Info
  series: dict[str, MetricInfo]
  media_preview: str | None = None
  size: float | None = None
  segments: dict[str, dict[str, Info]] | None = None

  def model_post_init(self, __context__) -> None:  # noqa: D105
    if not self.media_preview:
      self.media_preview = self.media_path
    self.info = dict(self.info)


class GraphInfo(TypedDict):
  adaptive_threshold: float
  period: dict[str, str]


class ClusterInfo(TypedDict):
  name: str


class NodeInfo(TypedDict):
  name: str
  label: str
  type: str
  image: str
  media_path: str
  cluster: int
  info: Info
  series: dict[str, MetricInfo]
  tags: list[dict[str, float]]
  segments: dict[str, str]


class CreativeMapJson(TypedDict):
  graph: GraphInfo
  clusters: dict[int, ClusterInfo]
  nodes: list[NodeInfo]
  edges: list[dict[str, int | float]]


class CreativeMap:
  """Defines CreativeMap based on a graph.

  Attributes:
    adaptive_threshold: Minimal value for defining similar media.
    fetching_request: Additional parameter used to generate a map.
    nodes: Information on each node of the map.
    edges: Information on each edge of the map.
    clusters: Aggregated information on each cluster of the map.
  """

  def __init__(
    self,
    adaptive_threshold: float,
    fetching_request: dict[str, Any] | None = None,
    nodes: list[NodeInfo] | None = None,
    edges: list[dict[str, int | float]] | None = None,
    clusters: dict[int, ClusterInfo] | None = None,
  ) -> None:
    """Initializes CreativeMap."""
    self.adaptive_threshold = adaptive_threshold
    self.fetching_request = fetching_request or {}
    self.nodes: list[NodeInfo] = nodes or []
    self.edges: list[dict[str, int | float]] = edges or []
    self.clusters: dict[int, ClusterInfo] = clusters or {}

  @classmethod
  def from_clustering(
    cls,
    clustering_results: media_similarity_service.ClusteringResults,
    tagging_results: Sequence[tagging_result.TaggingResult],
    extra_info: dict[str, MediaInfo] | None = None,
    fetching_request: dict[str, Any] | None = None,
    embed_previews: bool = False,
  ) -> CreativeMap:
    """Builds network visualization with injected extra_info."""
    tagging_mapping = {
      result.identifier: result.content for result in tagging_results
    }
    tagging_hash_to_identifier_mapping = {
      t.hash: t.identifier for t in tagging_results
    }
    media_type = tagging_results[0].type
    for node in clustering_results.graph.nodes:
      node_hash = node.get('name', '')
      node_name = tagging_hash_to_identifier_mapping.get(node_hash)
      if node_extra_info := extra_info.get(node_name):
        node['id'] = node_hash
        if size := node_extra_info.size:
          node['size'] = size
        node['type'] = media_type
        if media_type == 'text':
          node['image'] = _create_text_image_bytes(node_extra_info.media_path)
        elif embed_previews:
          node['image'] = _embed_preview(node_extra_info.media_preview)
        else:
          node['image'] = node_extra_info.media_preview
        node['media_path'] = node_extra_info.media_path
        node['label'] = node_extra_info.media_name or 'Unknown'
        node['cluster'] = clustering_results.clusters.get(node_name)
        node['info'] = node_extra_info.info
        node['series'] = node_extra_info.series
        node['tags'] = [
          {'tag': tag.name.replace("'", ''), 'score': tag.score}
          for tag in tagging_mapping.get(node_name, [])
        ]
        node['segments'] = node_extra_info.segments
    clusters = {
      cluster_id: f'Cluster: {cluster_id}'
      for cluster_id in set(clustering_results.clusters.values())
    }
    edges = [
      {
        'from': _from,
        'to': to,
        'similarity': similarity,
      }
      for _from, to, similarity in clustering_results.graph.edges
    ]
    return CreativeMap(
      adaptive_threshold=clustering_results.adaptive_threshold,
      fetching_request=fetching_request,
      nodes=clustering_results.graph.nodes,
      edges=edges,
      clusters=clusters,
    )

  def to_json(self) -> CreativeMapJson:
    """Extracts nodes from Network."""
    start_date = ''
    end_date = ''
    for node in self.nodes:
      if series := node.get('series'):
        min_start_date = min(series.keys())
        max_end_date = max(series.keys())
        if not start_date or min_start_date < start_date:
          start_date = min_start_date
        if not end_date or max_end_date > end_date:
          end_date = max_end_date

    return {
      'graph': {
        'adaptive_threshold': self.adaptive_threshold,
        'period': {
          'start_date': start_date or 'null',
          'end_date': end_date or 'null',
        },
      },
      'clusters': self.clusters,
      'nodes': self.nodes,
      'edges': self.edges,
    }

  def save(self, path: str | os.PathLike[str]) -> None:
    """Saves map to a json file."""
    with smart_open.open(path, 'w', encoding='utf-8') as f:
      json.dump(self.to_json(), f)


def convert_report_to_media_info(
  performance: report.GarfReport,
  media_type: media.MediaTypeEnum,
  metric_columns: Sequence[str] | None = None,
  segment_columns: Sequence[str] | None = None,
  modules: Sequence[str] | None = None,
  with_size_base: str | None = None,
) -> dict[str, MediaInfo]:
  """Convert report to MediaInfo mappings.

  Args:
    performance: Report with media data.
    media_type: Type of media in the report.
    metric_columns:  Name of metrics to be included in media info.
    segment_columns:  Name of segments to be calculated.
    modules: Optional column names to be added to media_info.
    with_size_base: Optional column name for regulating size of media info.

  Returns:
    Mapping between media identifier and its media info.
  """
  if isinstance(media_type, str):
    media_type = media.MediaTypeEnum[media_type.upper()]
  if with_size_base and with_size_base not in performance.column_names:
    logging.warning('Failed to set MediaInfo size to %s', with_size_base)
    with_size_base = None
  if with_size_base:
    try:
      float(performance[0][with_size_base])
    except TypeError:
      logging.warning('MediaInfo size attribute should be numeric')
    with_size_base = None

  if modules:
    modules = {module.split('.')[1] for module in modules}
    modules = modules.intersection(performance.column_names)
  else:
    modules = set()

  if media_type == media.MediaTypeEnum.YOUTUBE_VIDEO:
    modules.add('duration')
    modules.add('orientation')
  elif media_type == media.MediaTypeEnum.IMAGE:
    modules.add('format_type')

  metric_columns = metric_columns or []
  if 'in_campaigns' in performance.column_names:
    modules.add('in_campaigns')

  results = {}
  for media_url, values in performance.to_dict(key_column='media_url').items():
    info = build_info(values, list(metric_columns), list(modules))
    segments = (
      build_segments(values, segment_columns, metric_columns)
      if segment_columns
      else {}
    )
    if values[0].get('date'):
      series = build_segments(values, ['date'], metric_columns).get('date', {})
    else:
      series = {}
    if with_size_base and (size_base := info.get(with_size_base)):
      media_size = np.log(size_base) * np.log10(size_base)
    else:
      media_size = None
    results[media.convert_path_to_media_name(media_url, media_type)] = (
      MediaInfo(
        **create_node_links(media_url, media_type),
        media_name=values[0].get('media_name'),
        info=info,
        series=series,
        size=media_size,
        segments=segments,
      )
    )
  return results


def build_info(
  data: Info, metric_names: Sequence[str], modules: Sequence[str] | None = None
) -> Info:
  """Extracts and aggregated data for specified metrics.

  Args:
    data: All available information on media_url.
    metric_names: Metrics to sum.
    modules: Meterics/ dimensions to get as is without aggregation.

  Returns:
    Mapping between name of metric / dimension and its processed value.
  """
  info = {
    metric: _aggregate_nested_metric(data, metric) for metric in metric_names
  }
  if modules:
    dimensions = {module: data[0].get(module) for module in modules}
    info.update(dimensions)
  return info


def build_segments(
  data: Info, segment_names: Sequence[str], metric_names: Sequence[str]
) -> dict[str, dict[Info]]:
  """Builds info object for each variant of a segment.

  Args:
    data: Report data formatted as a mapping.
    segment_names: Names of column in report to transform into segments.
    metric_names: Names of metrics in report to calculate for each segment.

  Returns:
    Mapping between each segment, it's variants and corresponding metrics.
  """
  segments = {}
  for segment_name in segment_names:
    get_segment_getter = operator.itemgetter(segment_name)
    try:
      segment_values = set(map(get_segment_getter, data))
    except KeyError:
      continue
    segment_variants = {}
    for segment_value in segment_values:
      if segment_value != 'UNKNOWN':
        segment_variants[segment_value] = build_info(
          list(filter(lambda x: x[segment_name] == segment_value, data)),
          metric_names,
        )
    if segment_variants:
      segments[segment_name] = segment_variants
  return segments


def _aggregate_nested_metric(
  data: Info | Sequence[Info],
  metric_name: str,
) -> float | int | str | list[str] | None:
  """Performance appropriate aggregation over a dictionary.

  Sums numerical values and deduplicates and sorts alphabetically
  string values.

  Args:
    data: Data to extract metrics from.
    metric_name: Name of a metric to be extracted from supplied data.

  Returns:
    Aggregated value of a metric.
  """
  get_metric_getter = operator.itemgetter(metric_name)
  if isinstance(data, Mapping):
    return get_metric_getter(data)

  try:
    res = list(map(get_metric_getter, data))
  except KeyError:
    return None
  try:
    return sum(res)
  except TypeError:
    if len(result := sorted(set(res))) == 1:
      return ','.join(result)
    return result


def create_node_links(
  url: str, media_type: media.MediaTypeEnum
) -> dict[str, str]:
  return {
    'media_path': _to_youtube_video_link(url)
    if media_type == media.MediaTypeEnum.YOUTUBE_VIDEO
    else url,
    'media_preview': _to_youtube_preview_link(url)
    if media_type == media.MediaTypeEnum.YOUTUBE_VIDEO
    else url,
  }


def _to_youtube_preview_link(video_id: str) -> str:
  return f'https://img.youtube.com/vi/{video_id}/0.jpg'


def _to_youtube_video_link(video_id: str) -> str:
  return f'https://www.youtube.com/watch?v={video_id}'


def _create_text_image_bytes(
  text: str,
  format: str = 'PNG',
  font_size: int = 40,
  font_path: str | None = None,
  trim_treshold: int = 100,
  border=1,
):
  text = text[:trim_treshold]
  text = '\n'.join(textwrap.wrap(text, width=20))
  dummy_img = Image.new(mode='RGB', size=(1, 1))
  dummy_draw = ImageDraw.Draw(dummy_img)

  font = (
    ImageFont.truetype(font_path, font_size)
    if font_path
    else ImageFont.load_default(size=font_size)
  )

  bbox = dummy_draw.textbbox((0, 0), text, font=font)
  width = bbox[2] - bbox[0] + 10
  height = bbox[3] - bbox[1] + 10

  img = Image.new(mode='RGB', size=(width, height), color='white')
  img = ImageOps.expand(img, border=border, fill=(211, 211, 211))
  draw = ImageDraw.Draw(img)

  draw.multiline_text((5, 5), text, fill='black', font=font)

  byte_stream = BytesIO()
  img.save(byte_stream, format=format)

  image_bytes = byte_stream.getvalue()
  encoded_image = base64.b64encode(image_bytes).decode('utf-8')
  return f'data:image/png;base64,{encoded_image}'


def _embed_preview(url: str) -> str:
  logging.info('Embedding preview for url %s', url)
  with smart_open.open(url, 'rb') as f:
    image_preview = f.read()
  encoded_image = base64.b64encode(image_preview).decode('utf-8')
  return f'data:image/png;base64,{encoded_image}'



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""UI for generating creative map files."""

import datetime
import json
import os
import tempfile
from collections.abc import Sequence

import media_fetching
import media_similarity
import media_tagging
import smart_open
import streamlit as st
from pydantic_settings import BaseSettings

import filonov
from filonov import exceptions
from filonov.entrypoints import utils


class FilonovSettings(BaseSettings):
  """Specifies environmental variables for filonov.

  Ensure that mandatory variables are exposed via
  export ENV_VARIABLE_NAME=VALUE.

  Attributes:
    media_tagging_db_url: Connection string to DB with tagging results.
  """

  media_tagging_db_url: str | None = None


settings = FilonovSettings()


def streamlit_app():
  media_to_campaign_mapping = {
    'IMAGE': ['app', 'demandgen', 'pmax', 'display', 'search'],
    'TEXT': ['app', 'demandgen', 'pmax', 'display', 'search'],
    'YOUTUBE_VIDEO': ['app', 'demandgen', 'pmax'],
  }

  st.title('Creative Map Generator')

  source = st.selectbox(
    'Source', ['youtube', 'googleads', 'file', 'bq', 'sqldb', 'dbm']
  )

  with st.form('creative_map_form'):
    name = st.text_input('Name', 'creative_map')
    if source == 'googleads':
      tagger_type = st.selectbox('Tagger Type', ['gemini', 'google-cloud'])
      media_type = st.selectbox(
        'Media Type', ['IMAGE', 'YOUTUBE_VIDEO', 'TEXT']
      )
      account = st.text_input('Account', '')

      campaign_type = st.multiselect(
        'Campaign type', media_to_campaign_mapping.get(media_type)
      )

      col1, col2 = st.columns(2)
      with col1:
        start_date = st.date_input(
          'Start date: ',
          value=datetime.datetime.today() - datetime.timedelta(days=30),
        )
      with col2:
        end_date = st.date_input(
          'End date: ',
          value=datetime.datetime.today() - datetime.timedelta(days=1),
        )
        if end_date < start_date:
          st.error('End date cannot be less than start_date')
      extra_info = st.multiselect(
        'Extra info',
        [
          'googleads.main_geo',
          'googleads.approval_status',
          'tagging.language',
          'youtube.language',
        ],
      )
      input_parameters = {
        'account': account,
        'campaign_types': campaign_type,
        'start_date': start_date.strftime('%Y-%m-%d'),
        'end_date': end_date.strftime('%Y-%m-%d'),
        'extra_info': extra_info,
      }
      if not os.getenv('GOOGLE_ADS_CONFIGURATION_FILE_PATH'):
        st.error(
          'Path to google-ads.yaml is not found in '
          'GOOGLE_ADS_CONFIGURATION_FILE_PATH environmental variable. '
          'Please expose it or add manually.'
        )
        ads_config_path = st.text_input('Path to google-ads.yaml', '')
        input_parameters.update({'ads_config': ads_config_path})

    elif source in ('bq', 'file', 'sqldb'):
      input_parameters = {}
      if source == 'bq':
        table = st.text_input('bq_table', '')
        input_parameters['table'] = table
      elif source == 'sqldb':
        table = st.text_input('sql_table', '')
        connection_string = st.text_input('connection_string', '')
        input_parameters['table'] = table
        input_parameters['connection_string'] = connection_string
      elif source == 'file':
        file_data = st.file_uploader(label='Path to file', type='csv')
        tmp_file_path = None
        if file_data:
          with tempfile.NamedTemporaryFile(
            delete=False, mode='w', suffix='.csv', encoding='utf-8'
          ) as f:
            f.write(file_data.getvalue().decode('utf-8'))
            tmp_file_path = f.name
        input_parameters['path'] = tmp_file_path
      tagger_type = st.selectbox('Tagger Type', ['gemini', 'google-cloud'])
      media_type = st.selectbox(
        'Media Type', ['IMAGE', 'YOUTUBE_VIDEO', 'TEXT']
      )

      media_path_column = st.text_input('Media path column name', 'media_url')
      media_name = st.text_input('Media column name', 'media_name')
      metrics = st.text_input('Metric columns', 'clicks,impressions')
      segments = st.text_input('Segment columns', '')
      input_parameters.update(
        {
          'media_identifier': media_path_column,
          'media_name': media_name,
          'metrics': metrics.split(','),
          'segments': segments.split(','),
        }
      )
    elif source == 'youtube':
      media_type = 'YOUTUBE_VIDEO'
      tagger_type = 'gemini'
      channel = st.text_input('YouTube Channel ID', '')
      input_parameters = {'channel': channel}
      if not os.getenv('GARF_YOUTUBE_DATA_API_KEY'):
        st.error(
          'Add GARF_YOUTUBE_DATA_API_KEY environmental variable '
          'to work with YouTube.'
        )
    elif source == 'dbm':
      media_type = 'YOUTUBE_VIDEO'
      tagger_type = 'gemini'
      advertiser = st.text_input('DV 360 Advertiser ID', '')
      input_parameters = {'advertiser': advertiser}
      if not os.getenv('GARF_BID_MANAGER_CREDENTIALS_FILE'):
        st.error(
          'Add GARF_BID_MANAGER_CREDENTIALS_FILE environmental variable '
          'to work with Bid Manager.'
        )
    else:
      raise exceptions.FilonovError(f'Unsupported source: {source}')
    if not settings.media_tagging_db_url:
      st.warning(
        'No tagging data is saved between runs. '
        'Add MEDIA_TAGGING_DB_URL environmental variable to save data.'
      )

    col1, col2 = st.columns(2)
    with col1:
      submitted = st.form_submit_button('Generate Creative Map')
    with col2:
      cli_command_submitted = st.form_submit_button('Generate CLI command')

    if cli_command_submitted:
      request = filonov.CreativeMapGenerateRequest(
        source=source,
        media_type=media_type,
        tagger=tagger_type,
        source_parameters=input_parameters,
        output_parameters=filonov.filonov_service.OutputParameters(
          output_name=name
        ),
      )
      cli_command = build_cli_command(request, settings.media_tagging_db_url)
      st.code(cli_command, language='bash', wrap_lines=True)

    if submitted:
      request = filonov.CreativeMapGenerateRequest(
        source=source,
        media_type=media_type,
        tagger=tagger_type,
        source_parameters=input_parameters,
        output_parameters=filonov.filonov_service.OutputParameters(
          output_name=name
        ),
      )
      with st.expander('Show CLI command'):
        cli_command = build_cli_command(request, settings.media_tagging_db_url)
        st.code(cli_command, language='bash', wrap_lines=True)
      if 'fetching_service' not in st.session_state:
        st.session_state['fetching_service'] = (
          media_fetching.MediaFetchingService(source)
        )
      if 'tagging_service' not in st.session_state:
        st.session_state['tagging_service'] = media_tagging.MediaTaggingService(
          media_tagging.repositories.SqlAlchemyTaggingResultsRepository(
            settings.media_tagging_db_url
          )
        )
      if 'similarity_service' not in st.session_state:
        st.session_state['similarity_service'] = (
          media_similarity.MediaSimilarityService(
            media_similarity_repository=(
              media_similarity.repositories.SqlAlchemySimilarityPairsRepository(
                settings.media_tagging_db_url
              )
            )
          )
        )
      generated_map = filonov.FilonovService(
        st.session_state['fetching_service'],
        st.session_state['tagging_service'],
        st.session_state['similarity_service'],
      ).generate_creative_map(request)
      destination = utils.build_creative_map_destination(
        request.output_parameters.output_name
      )
      with smart_open.open(destination, 'w', encoding='utf-8') as f:
        json.dump(generated_map.to_json(), f)

      st.success(f'Creative Map saved to {destination}!')


def build_cli_command(
  request: filonov.CreativeMapGenerateRequest, db: str | None
) -> str:
  command_template = (
    'filonov --source {source} \\\n'
    '\t--media-type {media_type} \\\n'
    '\t--tagger {tagger} \\\n'
    '{source_parameters} \\\n'
    '\t--output-name {output}'
  )
  source = request.source.value
  source_parameters = []
  for name, value in request.source_parameters.model_dump().items():
    if not value or name == 'media_type':
      continue
    if isinstance(value, str):
      source_parameters.append(f'\t--{source}.{name}={value}')
    elif isinstance(value, Sequence):
      value_concat = ','.join(value)
      source_parameters.append(f'\t--{source}.{name}={value_concat}')

  source_parameters = ' \\\n'.join(source_parameters)
  params = {
    'source': source,
    'media_type': request.media_type,
    'tagger': request.tagger,
    'output': request.output_parameters.output_name,
    'source_parameters': source_parameters,
  }
  non_db_command = command_template.format(**params).strip()
  if db:
    return f'{non_db_command} \\\n\t--db-uri {db}'
  return non_db_command


streamlit_app()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CLI entrypoint for generating creative map."""

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

from typing import Optional

import media_fetching
import media_similarity
import media_tagging
import typer
from garf_executors.entrypoints import utils as garf_utils
from media_fetching.sources import models
from media_tagging import media
from media_tagging.entrypoints import utils as tagging_utils
from typing_extensions import Annotated

import filonov
from filonov.entrypoints import utils

typer_app = typer.Typer()


def _version_callback(show_version: bool) -> None:
  if show_version:
    print(f'filonov version: {media_fetching.__version__}')
    raise typer.Exit()


@typer_app.command(
  context_settings={'allow_extra_args': True, 'ignore_unknown_options': True}
)
@tagging_utils.log_shutdown
def main(
  ctx: typer.Context,
  source: Annotated[
    models.InputSource,
    typer.Option(
      help='Source to get data from',
      case_sensitive=False,
    ),
  ] = 'googleads',
  media_type: Annotated[
    media.MediaTypeEnum,
    typer.Option(
      help='Type of media',
      case_sensitive=False,
    ),
  ] = 'IMAGE',
  tagger: Annotated[
    Optional[str],
    typer.Option(
      help='Type of tagger',
      case_sensitive=False,
    ),
  ] = None,
  output_name: Annotated[
    str,
    typer.Option(
      help='Name of output file',
    ),
  ] = 'creative_map',
  db_uri: Annotated[
    Optional[str],
    typer.Option(
      help='Database connection string to store and retrieve results',
    ),
  ] = None,
  trim_tags_threshold: Annotated[
    Optional[float],
    typer.Option(
      help='Min allowed score for tags',
    ),
  ] = None,
  embed_previews: Annotated[
    bool,
    typer.Option(
      help='Whether media previews should be embedded into a creative map',
    ),
  ] = False,
  parallel_threshold: Annotated[
    int,
    typer.Option(
      help='Number of parallel processes to perform media tagging',
    ),
  ] = 10,
  logger: Annotated[
    garf_utils.LoggerEnum,
    typer.Option(
      help='Type of logger',
    ),
  ] = 'rich',
  loglevel: Annotated[
    str,
    typer.Option(
      help='Level of logging',
    ),
  ] = 'INFO',
  log_name: Annotated[
    str,
    typer.Option(
      help='Name of logger',
    ),
  ] = 'filonov',
  version: Annotated[
    bool,
    typer.Option(
      help='Display library version',
      callback=_version_callback,
      is_eager=True,
      expose_value=False,
    ),
  ] = False,
):  # noqa: D103
  garf_utils.init_logging(
    loglevel=loglevel.upper(), logger_type=logger, name=log_name
  )
  supported_enrichers = (
    media_fetching.enrichers.enricher.AVAILABLE_MODULES.keys()
  )
  parsed_param_keys = set(
    [source, 'tagger', 'similarity'] + list(supported_enrichers)
  )
  extra_parameters = garf_utils.ParamsParser(parsed_param_keys).parse(ctx.args)
  fetching_service = media_fetching.MediaFetchingService.from_source_alias(
    source
  )
  tagging_service = media_tagging.MediaTaggingService(
    tagging_results_repository=(
      media_tagging.repositories.SqlAlchemyTaggingResultsRepository(db_uri)
    )
  )
  similarity_service = media_similarity.MediaSimilarityService(
    media_similarity.repositories.SqlAlchemySimilarityPairsRepository(db_uri)
  )
  if source == 'youtube':
    media_type = 'YOUTUBE_VIDEO'
    tagger = 'gemini'
  request = filonov.CreativeMapGenerateRequest(
    source=source,
    media_type=media_type,
    tagger=tagger,
    tagger_parameters=extra_parameters.get('tagger'),
    similarity_parameters=extra_parameters.get('similarity'),
    source_parameters=extra_parameters.get(source),
    output_parameters=filonov.filonov_service.OutputParameters(
      output_name=output_name
    ),
    parallel_threshold=parallel_threshold,
    trim_tags_threshold=trim_tags_threshold,
    embed_previews=embed_previews,
    context=extra_parameters,
  )
  filonov_service = filonov.FilonovService(
    fetching_service, tagging_service, similarity_service
  )
  generated_map = filonov_service.generate_creative_map(request)
  destination = utils.build_creative_map_destination(
    request.output_parameters.output_name
  )
  generated_map.save(destination)


if __name__ == '__main__':
  typer_app()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides HTTP endpoint for filonov requests."""

# pylint: disable=C0330, g-bad-import-order, g-multiple-import
from typing import Literal

import fastapi
import media_fetching
import media_similarity
import media_tagging
import uvicorn
from media_similarity.entrypoints.server import (
  router as media_similarity_router,
)
from media_tagging.entrypoints.server import router as media_tagging_router
from pydantic_settings import BaseSettings
from typing_extensions import Annotated

import filonov
from filonov.entrypoints import utils


class FilonovSettings(BaseSettings):
  """Specifies environmental variables for filonov.

  Ensure that mandatory variables are exposed via
  export ENV_VARIABLE_NAME=VALUE.

  Attributes:
    media_tagging_db_url: Connection string to DB with tagging results.
  """

  media_tagging_db_url: str | None = None


class Dependencies:
  def __init__(self) -> None:
    """Initializes CommonDependencies."""
    settings = FilonovSettings()
    self.tagging_service = media_tagging.MediaTaggingService(
      media_tagging.repositories.SqlAlchemyTaggingResultsRepository(
        settings.media_tagging_db_url
      )
    )
    self.similarity_service = media_similarity.MediaSimilarityService(
      media_similarity_repository=(
        media_similarity.repositories.SqlAlchemySimilarityPairsRepository(
          settings.media_tagging_db_url
        )
      ),
    )


router = fastapi.APIRouter(prefix='/creative_maps')


class CreativeMapGoogleAdsGenerateRequest(filonov.CreativeMapGenerateRequest):
  """Specifies Google Ads specific request for returning creative map."""

  source_parameters: (
    media_fetching.sources.googleads.GoogleAdsFetchingParameters
  )
  source: Literal['googleads'] = 'googleads'


class CreativeMapFileGenerateRequest(filonov.CreativeMapGenerateRequest):
  """Specifies Google Ads specific request for returning creative map."""

  source_parameters: media_fetching.sources.file.FileFetchingParameters
  source: Literal['file'] = 'file'


class CreativeMapYouTubeGenerateRequest(filonov.CreativeMapGenerateRequest):
  """Specifies YouTube specific request for returning creative map."""

  source_parameters: media_fetching.sources.youtube.YouTubeFetchingParameters
  source: Literal['youtube'] = 'youtube'
  media_type: Literal['YOUTUBE_VIDEO'] = 'YOUTUBE_VIDEO'
  tagger: Literal['gemini'] = 'gemini'


@router.post('/generate:file')
async def generate_creative_map_file(
  request: CreativeMapFileGenerateRequest,
  dependencies: Annotated[Dependencies, fastapi.Depends(Dependencies)],
) -> fastapi.responses.JSONResponse:
  """Generates Json with creative map data."""
  return generate_creative_map(
    'file',
    request,
    dependencies,
  )


@router.post('/generate:googleads')
async def generate_creative_map_googleads(
  request: CreativeMapGoogleAdsGenerateRequest,
  dependencies: Annotated[Dependencies, fastapi.Depends(Dependencies)],
) -> fastapi.responses.JSONResponse:
  """Generates Json with creative map data."""
  return generate_creative_map(
    'googleads',
    request,
    dependencies,
  )


@router.post('/generate:youtube')
async def generate_creative_map_youtube(
  request: CreativeMapYouTubeGenerateRequest,
  dependencies: Annotated[Dependencies, fastapi.Depends(Dependencies)],
) -> fastapi.responses.JSONResponse:
  """Generates Json with creative map data."""
  return generate_creative_map(
    'youtube',
    request,
    dependencies,
  )


def generate_creative_map(
  source: Literal['youtube', 'googleads', 'file'],
  request: filonov.CreativeMapGenerateRequest,
  dependencies: Annotated[Dependencies, fastapi.Depends(Dependencies)],
) -> filonov.creative_map.CreativeMapJson:
  """Generates Json with creative map data."""
  generated_map = (
    filonov.FilonovService(
      fetching_service=media_fetching.MediaFetchingService(source),
      tagging_service=dependencies.tagging_service,
      similarity_service=dependencies.similarity_service,
    )
    .generate_creative_map(request)
    .to_json()
  )

  if request.output_parameters.output_type == 'file':
    destination = utils.build_creative_map_destination(
      request.output_parameters.output_name
    )
    generated_map.save(destination)
    return fastapi.responses.JSONResponse(
      content=f'Creative map was saved to {destination}.'
    )

  return fastapi.responses.JSONResponse(
    content=fastapi.encoders.jsonable_encoder(generated_map)
  )


app = fastapi.FastAPI()
app.include_router(router)
app.include_router(media_tagging_router)
app.include_router(media_similarity_router)


def main():
  uvicorn.run(app)


if __name__ == '__main__':
  main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Handles creative map generation."""

import logging
from collections.abc import Sequence
from typing import Any, ClassVar, Literal

import media_fetching
import media_similarity
import media_tagging
import pydantic
from media_tagging import media
from media_tagging.media_tagging_service import (
  MediaFetchingRequest,
  MediaTaggingRequest,
)

from filonov import creative_map, exceptions

logger = logging.getLogger('filonov')


class SimilarityParameters(pydantic.BaseModel):
  normalize: bool = False
  custom_threshold: float | None = None
  algorithm: str | None = None


class OutputParameters(pydantic.BaseModel):
  """Parameters for saving creative maps data.

  Attributes:
    output_name: Fully qualified name to store the results.
    output_type: Type of output.

  """

  output_name: str = 'creative_map'
  output_type: Literal['console', 'file'] = 'file'


class CreativeMapGenerateRequest(pydantic.BaseModel):
  """Specifies structure of request for returning creative map.

  Attributes:
    source: Source of getting data for creative map.
    media_type: Type of media to get.
    tagger: Type of tagger to use.
    tagger_parameters: Parameters to finetune tagging.
    similarity_parameters: Parameters to similarity matching.
    source_parameters: Parameters to get data from the source of creative map.
    output_parameters: Parameters for saving creative maps data.
    parallel_threshold: Tagging and similarity detecting threshold.
    trim_tags_threshold: Keeps tags only with the score higher than threshold.
    embed_previews: Whether media previews should be embedded into a map.
    context: Overall context of map generation.
  """

  default_tagger_parameters: ClassVar[dict[str, int]] = {'n_tags': 100}

  source: media_fetching.sources.models.InputSource
  media_type: Literal[tuple(media.MediaTypeEnum.options())] | None = None
  tagger: Literal['gemini', 'google-cloud', 'loader', None] = None
  tagger_parameters: dict[str, str | int] = default_tagger_parameters
  similarity_parameters: SimilarityParameters = SimilarityParameters()
  source_parameters: dict[str, str | Sequence[str]] = pydantic.Field(
    default_factory=dict
  )
  output_parameters: OutputParameters = OutputParameters()
  parallel_threshold: int = 10
  trim_tags_threshold: float | None = None
  embed_previews: bool = False
  context: dict[str, Any] = pydantic.Field(default_factory=dict)

  def model_post_init(self, __context):  # noqa: D105
    if not self.tagger_parameters:
      self.tagger_parameters = self.default_tagger_parameters
    if 'n_tags' not in self.tagger_parameters:
      self.tagger_parameters.update(self.default_tagger_parameters)
    if self.source == 'youtube':
      self.media_type = 'YOUTUBE_VIDEO'
      self.tagger = 'gemini'

    source_parameters_class = media_fetching.INPUT_MAPPING.get(self.source)
    if not isinstance(self.source_parameters, source_parameters_class):
      self.source_parameters = source_parameters_class(
        **self.source_parameters, media_type=self.media_type
      )

    self.context.update({self.source: self.source_parameters.model_dump()})


class FilonovService:
  """Responsible for handling requests for creative map input generation."""

  def __init__(
    self,
    fetching_service: media_fetching.MediaFetchingService | None = None,
    tagging_service: media_tagging.MediaTaggingService | None = None,
    similarity_service: media_similarity.MediaSimilarityService | None = None,
  ) -> None:
    """Initializes FilonovService."""
    self.fetching_service = fetching_service
    self.tagging_service = tagging_service
    self.similarity_service = similarity_service

  def generate_creative_map(
    self,
    request: CreativeMapGenerateRequest,
  ) -> creative_map.CreativeMap:
    """Generates creative map data.

    Performs the following steps:

    * Input data fetching.
    * Media tagging.
    * Media similarity matching.

    Args:
      request: Request for creative maps generation.

    Returns:
      Generated creative map.

    Raises:
      FilonovError: When performance or tagging data not found.
    """
    if not request.tagger and not self.tagging_service:
      raise exceptions.FilonovError(
        'Failed to get tagging results from DB. MediaTaggingService missing.'
      )
    if not (fetching_service := self.fetching_service):
      fetching_service = media_fetching.MediaFetchingService(request.source)
    media_data = fetching_service.fetch(
      request.source_parameters, request.context
    )
    if not media_data:
      raise exceptions.FilonovError(
        'No performance data found for the context.'
      )
    media_info = creative_map.convert_report_to_media_info(
      performance=media_data,
      media_type=request.media_type,
      metric_columns=request.source_parameters.metrics,
      segment_columns=request.source_parameters.segments,
      modules=request.source_parameters.extra_info,
    )
    media_urls = {media.media_path for media in media_info.values()}
    if not (tagging_service := self.tagging_service):
      tagging_service = media_tagging.MediaTaggingService()
    if not request.tagger:
      logger.info('Tagger not specified, getting data from DB')
      tagging_response = tagging_service.get_media(
        MediaFetchingRequest(
          media_type=request.media_type,
          media_paths=list(media_urls),
          output='tag',
          deduplicate=True,
        )
      )
      if not tagging_response:
        raise exceptions.FilonovError('No tagging data found in DB.')
    else:
      tagging_response = tagging_service.tag_media(
        MediaTaggingRequest(
          tagger_type=request.tagger,
          media_type=request.media_type,
          tagging_options=request.tagger_parameters,
          media_paths=media_urls,
          parallel_threshold=request.parallel_threshold,
          deduplicate=True,
        ),
        path_processor=request.tagger_parameters.get('path_processor'),
      )
    if not tagging_response:
      raise exceptions.FilonovError(
        'Failed to perform media tagging for the context: '
        f'{request.source_parameters}'
      )
    if not (similarity_service := self.similarity_service):
      similarity_service = (
        media_similarity.MediaSimilarityService.from_connection_string(
          tagging_service.repo.db_url
        )
      )
    clustering_request = media_similarity.MediaClusteringRequest(
      media_type=request.media_type,
      tagger_type=request.tagger,
      tagging_options=request.tagger_parameters,
      normalize=request.similarity_parameters.normalize,
      custom_threshold=request.similarity_parameters.custom_threshold,
      algorithm=request.similarity_parameters.algorithm,
      parallel_threshold=request.parallel_threshold,
      tagging_response=tagging_response,
    )
    clustering_results = similarity_service.cluster_media(clustering_request)
    logger.info('Generating creative map...')
    if trim_threshold := request.trim_tags_threshold:
      tagging_response.trim(trim_threshold)
    return creative_map.CreativeMap.from_clustering(
      clustering_results=clustering_results,
      tagging_results=tagging_response.results,
      extra_info=media_info,
      fetching_request=request.source_parameters.model_dump(),
      embed_previews=request.embed_previews,
    )
